/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/* these headers are needed by all files */
/*lint -e129 these resolve several issues pclint has with my system headers */
/*lint -e78 */
/*lint -e830 */
/*lint -e38 */
/*lint -e148 */
/*lint -e114 */
/*lint -e18 */
/*lint -e49 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <time.h>
#include <string>
#include <vector>
#include <valarray>
#include <complex>
#ifdef DMALLOC
#include <dmalloc.h>
#endif
/*lint +e18 */
/*lint +e49 */
/*lint +e38 */
/*lint +e148 */
/*lint +e830 */
/*lint +e78 */
/*lint -e129 */

using namespace std;

/** define EXTERN as nothing before including cddefines.h to allocate global data */
#ifndef EXTERN
#define EXTERN extern 
#endif

#undef STATIC
#ifdef USE_GPROF
#define STATIC
#else
#define STATIC static
#endif

/* make sure this is globally visible as well! */
#include "cpu.h"


//*************************************************************************
//
//! Singleton: template to construct classes where a single instance can
//! be accessed across the entire program. Construct your class as follows:
//!
//! class t_myclass : public Singleton<t_myclass>
//! {
//! 	friend class Singleton<t_myclass>;
//! protected:
//! 	t_myclass(); // make sure the contructor is protected!
//! public:
//! 	long myfunc() const { return 43; }
//! };
//!
//! and use as follows throughout the code:
//!
//! long l = t_myclass::Inst().myfunc();
//!
//!   NB NB - This implementation is not threadsafe !!
//
// This implementation has been obtained from Wikipedia
//
//*************************************************************************

template<typename T> class Singleton
{
public:
	static T& Inst()
	{
		static T instance__;  // assumes T has a protected default constructor
		return instance__;
	}
};


/**************************************************************************
 *
 * these are variables and pointers for output from the code, used everywhere
 * declared extern here, and definition is in cddefines.cpp
 *
 **************************************************************************/

/** ioQQQ is the file hand to the output file itself,
 * ioQQQ is set to stdout by default in cdInit, 
 * and is reset to anything else by calling cdOutp */
EXTERN FILE *ioQQQ;

EXTERN FILE *ioStdin;

extern FILE *ioMAP;

/** we shall write errors to this file, it is set
 * to stderr in cdInit */
EXTERN FILE* ioPrnErr;

/** this set true when abort sequence is initiated - serious meltdown is happening */
EXTERN bool lgAbort;

/** flag lgTestIt turned on if routine TestCode ever called, only generates
 * comment that test code is in place */
EXTERN bool lgTestCodeCalled; 

/** flag lgTestOn set true with SET TEST command
 * for some test code to be run somewhere */
EXTERN bool lgTestCodeEnabled;

/** this is flag saying whether to print errors to 
 * standard error output */
EXTERN bool lgPrnErr;

/** this tells the code to throw an fpe when an assert is thrown,
 * set true with SET ASSERT FPE command */
EXTERN bool lgAssertFPE;

/** nzone is zone counter, incremented in routine cloudy 
 * is zero during search phase, 1 for first zone at illuminated face */
EXTERN long int nzone;

/** this is nzone + conv.nPres2Ioniz/100 in ConvBase */
EXTERN double fnzone;

/** the iteration counter, set and incremented in routine cloudy,
 * ==1 during first iteration, 2 during second, etc */
EXTERN long int iteration;

/**
 * this is the number zero, used to trick clever compilers when 
 * dividing by it to crash program
 * there is a routine called zero - this name cannot overlap
 * definition is in cddefines.cpp
 */
extern const double ZeroNum;


/**************************************************************************
 *
 * these are constants used to dimension several vectors and index arrays
 *
 **************************************************************************/

/** FILENAME_PATH_LENGTH is the size of the string that holds the path.  The longest
 * string that can be held is one less than this, due to the end
 * end of string sentinel in C.  Increase this is a larger string
 * is needed to hold the path on your system */
const int FILENAME_PATH_LENGTH = 200; 

/** twice the above, so that we can add file name to end of full path */
const int FILENAME_PATH_LENGTH_2  = FILENAME_PATH_LENGTH*2; 

/** this is limit to longest line of information that is scanned in,
 * end of line char is actually at this +1, dim of vector is this + 1 
 * all routines that scan in information should use this to dim vars */ 
const int INPUT_LINE_LENGTH = 200;

/** This is the number of elements included in the code,
 * is used to set lengths of many vectors */
const int LIMELM = 30;

/** the number of iso sequences now in the code */
const int NISO = 2;

/** following is real limit to how many levels can be computed for
 * model hydrogen atom - this has the one extra included, so at most
 * iso.numLevels_max[ipH_LIKE] can be NHYDRO_MAX_LEVEL-1 and vectors should be dim NHYDRO_MAX_LEVEL*/
const int NHYDRO_MAX_LEVEL = 401;

/** dimension of vector to do with number of iterations stored
 * so, this is the the most that can possibly be performed */
/*const int ITR DIM = 200;*/

/** number of hydrogen species in hmole */
const int N_H_MOLEC = 8;

/** stopping temperature */
const double TEND = 4000.;

/** this is used to add to depth to prevent div or log of zero */
const double DEPTH_OFFSET = 1.e-30;

 /* these are flags for various colliders that are used across the code */
const int ipELECTRON = 0;
const int ipPROTON = 1;
const int ipHE_PLUS = 2;

/* indices within recombination coefficient array */
/* ipRecEsc is state specific escape probability*/
const int ipRecEsc = 2;
/* the net escaping, including destruction by background and optical deepth*/
const int ipRecNetEsc = 1;
/* ipRecRad is state specific radiative recombination rate*/
const int ipRecRad = 0;
/** with the above, the total radiative rec per ion is
 * iso.RadRecomb[ipISO][nelem][n][ipRecRad]*
   iso.RadRecomb[ipISO][nelem][n][ipRecNetEsc]*dense.eden; */

/* these specify the form of the line redistribution function */
/* partial redistribution with wings */
const int ipPRD = 1;
/* complete redistribution, core only, no wings, Hummer's K2 function */
const int ipCRD = -1;
/* complete redistribution with wings */
const int ipCRDW = 2;
/* redistribution function for Lya, calls Hummer routine for H-like series only */
const int ipLY_A = -2;
/* core function for K2 destruction */
const int ipDEST_K2 = 1;
/* core function for complete redist destruction */
const int ipDEST_INCOM = 2;
/* core function for simple destruction */
const int ipDEST_SIMPL = 3;

/** some levels for hydrogenic species */

const int ipH1s = 0;
const int ipH2s = 1;
const int ipH2p = 2;

/** some levels for he-like species */

/* level 1 */
const int ipHe1s1S = 0;

/* level 2 */
const int ipHe2s3S = 1;
const int ipHe2s1S = 2;
const int ipHe2p3P0 = 3;
const int ipHe2p3P1 = 4;
const int ipHe2p3P2 = 5;
const int ipHe2p1P = 6;

/* level 3 */
const int ipHe3s3S = 7;
const int ipHe3s1S = 8;
const int ipHe3p3P = 9;
const int ipHe3d3D = 10;
const int ipHe3d1D = 11;
const int ipHe3p1P = 12;

/* level 4 */
const int ipHe4s3S = 13;
const int ipHe4s1S = 14;
const int ipHe4p3P = 15;
const int ipHe4d3D = 16;
const int ipHe4d1D = 17;
const int ipHe4f3F = 18;
const int ipHe4f1F = 19;
const int ipHe4p1P = 20;

/* level 5 */
const int ipHe5s3S = 21;
const int ipHe5s1S = 22;
const int ipHe5p3P = 23;
const int ipHe5d3D = 24;
const int ipHe5d1D = 25;
const int ipHe5f3F = 26;
const int ipHe5f1F = 27;
const int ipHe5g3G = 28;
const int ipHe5g1G = 29;
const int ipHe5p1P = 30;

/** these are array indices for isoelectronic sequences,
 * same as element but used for array addressing to make
 * context totally clear */
const int ipH_LIKE = 0;
const int ipHE_LIKE = 1;
const int ipLI_LIKE = 2;
const int ipBE_LIKE = 3;
const int ipB_LIKE = 4;
const int ipC_LIKE = 5;
const int ipN_LIKE = 6;
const int ipO_LIKE = 7;
const int ipF_LIKE = 8;

/** these are incides to some elements, on the C scale */
const int ipHYDROGEN = 0;
const int ipHELIUM = 1;
const int ipLITHIUM = 2;
const int ipBERYLLIUM = 3;
const int ipBORON = 4;
const int ipCARBON = 5;
const int ipNITROGEN = 6;
const int ipOXYGEN = 7;
const int ipFLUORINE = 8;
const int ipNEON = 9;
const int ipSODIUM = 10;
const int ipMAGNESIUM = 11;
const int ipALUMINIUM = 12;
const int ipSILICON = 13;
const int ipPHOSPHORUS = 14;
const int ipSULPHUR = 15;
const int ipCHLORINE = 16;
const int ipARGON = 17;
const int ipPOTASSIUM = 18;
const int ipCALCIUM = 19;
const int ipSCANDIUM = 20;
const int ipTITANIUM = 21;
const int ipVANADIUM = 22;
const int ipCHROMIUM = 23;
const int ipMANGANESE = 24;
const int ipIRON = 25;
const int ipCOBALT = 26;
const int ipNICKEL = 27;
const int ipCOPPER = 28;
const int ipZINC = 29;


/**************************************************************************
 *
 * various macros used by the code
 *
 **************************************************************************/

/** to avoid errors introduced by C's infamous double-negative logic,
 * this uses NDEBUG (the ANSI std macro used to tell assert that
 * we are not debugging) to define DEBUG */
#ifndef NDEBUG
#	define DEBUG
#else
#	undef DEBUG
#endif

/** as the code is developed this should become part of the design for future
 * true parallel processing.  This flag should be false if it is not necessary
 * for a given instance of a routine to keep separate memory.  This is the case in
 * a serial machine, or an MPI run where there are n instances of the code running
 * on n processors.  This should be true if an instance of a routine must have its
 * own unique memory, so that other instances of the routine, which may be running
 * in parallel on another processor, does not share the same memory.  */
#ifndef PARALLEL_MODE
#define PARALLEL_MODE false
#endif

/** use special version of malloc - it tests result and dies if cannot allocate space */
#if defined(malloc)
/* ...but if malloc is a macro, assume it is instrumented by a memory debugging tool
 * (e.g. dmalloc) */
#	define MALLOC(exp) (malloc(exp))
#else
/* Otherwise instrument and protect it ourselves */
#	define MALLOC(exp) (MyMalloc(exp,__FILE__, __LINE__))
#endif

/** now special version of calloc - it dies if cannot allocate space. */
#if defined(calloc)
/* ...but if calloc is a macro, assume it is instrumented by a memory debugging tool */
#	define CALLOC calloc
#else
/* Otherwise instrument and protect it ourselves */
#	define CALLOC MyCalloc 
#endif

/** now special version of calloc - it dies if cannot allocate space. */
#if defined(realloc)
/* ...but if calloc is a macro, assume it is instrumented by a memory debugging tool */
#	define REALLOC realloc
#else
/* Otherwise instrument and protect it ourselves */
#	define REALLOC MyRealloc 
#endif

/** a cloudy version of the assert macro */
#undef  ASSERT
#ifndef STD_ASSERT
#	ifdef NDEBUG
#		define ASSERT(exp) ((void)0)
#	else
/* a version of assert that fails gracefully */
#		define ASSERT(exp) if (!(exp)) \
		MyAssert(__FILE__, __LINE__)
#	endif
#else
#	define ASSERT(exp)	(assert(exp))
#endif

/* Windows does not define isnan */
/* use our version on all platforms since the isnanf
 * function does not exist under Solaris 9 either */
#undef isnan
#define isnan MyIsnan

/** entry and exit of each routine will go here,
 * macros enabled if compiler-set flag DEBUG_FUN is defined */
class t_debugprt : public Singleton<t_debugprt>
{
	friend class Singleton<t_debugprt>;
protected:
	t_debugprt()
	{
		indent = 0;
		debug_fp = stderr;
	}
private:
	int indent;
	FILE *debug_fp;
	void print_str( const char *str, char dir ) const
	{
		for( int i=0; i < indent; i++ )
			fputc( ' ', debug_fp );
		fputc( dir, debug_fp );
		fputs( str, debug_fp );
		fputc( '\n', debug_fp );
	}
public:
	void printin( const char *str )
	{
		print_str( str, '>' );
		indent++;
	}
	void printout( const char *str )
	{
		--indent;
		print_str( str, '<' );
	}
};

#ifdef DEBUG_FUN
#define DEBUG_ENTRY( STR ) t_debugprt::Inst().printin( STR )
#define DEBUG_EXIT( STR ) t_debugprt::Inst().printout( STR )
#else
#define DEBUG_ENTRY( STR ) ((void)0)
#define DEBUG_EXIT( STR ) ((void)0)
#endif

/* TorF(l) returns a 'T' or 'F' depending on the 'logical' expr 'l' */
inline char TorF( bool l ) { return l ? 'T' : 'F'; }
/* */

/** checks whether argument is odd */
inline bool is_odd( int j ) { return (j&1) == 1; }
inline bool is_odd( long j ) { return (j&1L) == 1L; }
/* */

/** nint rounds to the nearest long int */
inline long nint( double x ) { return static_cast<long>( (x < 0.) ? x-0.5 : x+0.5 ); }
/* */

/* define min for mixed arguments, the rest already exists */
inline long min( int a, long b ) { long c = a; return ( (c < b) ? c : b ); }
inline long min( long a, int b ) { long c = b; return ( (a < c) ? a : c ); }
inline double min( float a, double b ) { double c = a; return ( (c < b) ? c : b ); }
inline double min( double a, float b ) { double c = b; return ( (a < c) ? a : c ); }

#undef MIN2
/** MIN2 takes two arguments, returns the smaller of the two */
// #define MIN2(a,b) (((a)<(b)) ? (a) : (b))
#define MIN2 min
/* */

#undef MIN3
/** MIN3 takes 3 arguments, returns the smallest of the 3 */
#define MIN3(a,b,c) (MIN2(MIN2(a,b),c))
/* */

#undef MIN4
/** MIN2 takes 4 arguments, returns the smallest of the 4 */
#define MIN4(a,b,c,d) (MIN2(MIN2(a,b),MIN2(c,d)))
/* */

/* define max for mixed arguments, the rest already exists */
inline long max( int a, long b ) { long c = a; return ( (c > b) ? c : b ); }
inline long max( long a, int b ) { long c = b; return ( (a > c) ? a : c ); }
inline double max( float a, double b ) { double c = a; return ( (c > b) ? c : b ); }
inline double max( double a, float b ) { double c = b; return ( (a > c) ? a : c ); }

#undef MAX2
/** MAX2 takes two arguments, returns the larger of the two */
// #define MAX2(a,b) (((a)>(b)) ? (a) : (b))
#define MAX2 max
/* */

#undef MAX3
/** MAX3 takes 3 arguments, returns the largest of the 3 */
#define MAX3(a,b,c) (MAX2(MAX2(a,b),c))
/* */

#undef MAX4
/** MAX2 takes 4 arguments, returns the largest of the 4 */
#define MAX4(a,b,c,d) (MAX2(MAX2(a,b),MAX2(c,d)))
/* */

/** floating point sign transfer (fortran sign function) - sign of y times abs value of x 
\param x
\param y
*/
template<class T>
inline T sign( T x, T y )
{
	return ( y < T() ) ? -abs(x) : abs(x);
}
/* */

/** sign3 returns -1 for negative arguments, +1 for positive, and 0 for zero (pascal sign function) */
template<class T>
inline int sign3( T x ) { return ( x < T() ) ? -1 : ( ( x > T() ) ? 1 : 0 ); }
/* */

/** checks whether two FP numbers are "equal" (differ no more than 3 epsilon) */
inline bool fp_equal( float x, float y )
{
	// mimic IEEE behavior
	if( isnan(x) || isnan(y) )
		return false;
	int sx = sign3(x);
	int sy = sign3(y);
	// treat zero cases first to avoid division by zero below
	if( sx == 0 && sy == 0 )
		return true;
	// either x or y is zero (but not both), or x and y have different sign
	if( sx*sy != 1 )
		return false;
	x = abs(x);
	y = abs(y);
	return ( 1.f - min(x,y)/max(x,y) < 3.1f*FLT_EPSILON );
}

inline bool fp_equal( double x, double y )
{
	// mimic IEEE behavior
	if( isnan(x) || isnan(y) )
		return false;
	int sx = sign3(x);
	int sy = sign3(y);
	// treat zero cases first to avoid division by zero below
	if( sx == 0 && sy == 0 )
		return true;
	// either x or y is zero (but not both), or x and y have different sign
	if( sx*sy != 1 )
		return false;
	x = abs(x);
	y = abs(y);
	return ( 1. - min(x,y)/max(x,y) < 3.1*DBL_EPSILON );
}

#undef POW2
/** POW2 takes 1 argument, and squares it */
// #define POW2(a) ((a)*(a))
#define POW2 pow2
inline int pow2(int a) { return a*a; }
inline long pow2(long a) { return a*a; }
inline float pow2(float a) { return a*a; }
inline double pow2(double a) { return a*a; }
/* */

#undef POW3
/** POW3 takes 1 argument, and cubes it */
// #define POW3(a) ((a)*(a)*(a))
#define POW3 pow3
inline int pow3(int a) { return a*a*a; }
inline long pow3(long a) { return a*a*a; }
inline float pow3(float a) { return a*a*a; }
inline double pow3(double a) { return a*a*a; }
/* */

#undef POW4
/** POW4 takes 1 argument, and raises it to the power 4 */
#define POW4(X) (POW2(X)*POW2(X))
/* */

#undef SDIV
/** SDIV safe division - div by SDIV(x) - if abs val of arg >SMALLFLOAT,
 * returns arg, if < SMALLFLOAT, returns SMALLFLOAT - with negative arg
 * return is +SMALLFLOAT so sign changes */
// #define SDIV(b) (((SMALLFLOAT)>(fabs(b))) ? (SMALLFLOAT) : (b))
inline float SDIV( float x ) { return ( fabs((double)x) < (double)SMALLFLOAT ) ? SMALLFLOAT : x; }
/* \todo should we use SMALLDOUBLE here ? it produces overflows now... PvH */
inline double SDIV( double x ) { return ( fabs(x) < (double)SMALLFLOAT ) ? (double)SMALLFLOAT : x; }
// inline double SDIV( double x ) { return ( fabs(x) < SMALLDOUBLE ) ? SMALLDOUBLE : x; }
/* */

/** macro to determine size of an array -
 * from p 22, Practice of Programming, 
 * by B.W. Kernighan & Rob Pike, Addison Wesley.
 * use: 
 * double dbug[100];
 for( i=0; i<NELEMS(dbug); i++)
  ....
 */
//   unused....
// #undef NELEMS
// #define NELEMS(array) (sizeof(array)/sizeof(array[0]))

// unused.... what is this for ???
#ifndef _ERR_
#define _ERR_	-1
#endif
/* */

#undef HMRATE
/*HMRATE compile molecular rates using Hollenbach and McKee fits */
/* #define HMRATE(a,b,c) ( ((b) == 0 && (c) == 0) ? (a) : \
 *	( ((c) == 0) ? (a)*pow(phycon.te/300.,(b)) : \
 *	( ((c)/phycon.te > 50.) ? 0. : ( ((b) == 0) ?  (a)*exp(-(c)/phycon.te) : \
 *					 (a)*pow(phycon.te/300.,(b))*exp(-(c)/phycon.te) ) ) ) ) */
#define HMRATE(a,b,c) hmrate4(a,b,c,phycon.te)

inline double hmrate4( double a, double b, double c, double te )
{
	if( b == 0. && c == 0. )
		return a;
	else if( c == 0. )
		return a*pow(te/300.,b);
	else if( b == 0. )
		return ( c/te <= 50. ) ? a*exp(-c/te) : 0.;
	else
		return ( c/te <= 50. ) ? a*pow(te/300.,b)*exp(-c/te) : 0.;
}


/********************************************************************
 *
 * this structure defines the main emission line class.  
 *
 *********************************************************************/
typedef struct 
{
	/**< all of these are set to dangerous values by EmLineJunk */

	/** type of redistribution function, 
	-1 complete redis Dopper core only, 
	 0 insanity
	 1 incomplete redistirbution with damping wings, 
	 2 complete redistribution with damping wings,
	*/
	int iRedisFun;

	/** index for line within continuum array,
	 * this is on the f, not c, scale,
	 * negative ipCont means this is not a real line
	 * sentinal for fake line */
	int ipCont;

	/** index for line within fine continuum array, for line overlap */
	long int ipFine;

	 /** ion stage of element, 1 for atom, 2 ion, etc */
	int IonStg;

	 /** atomic number of element, 1 for H, 2 for He, etc */
	int nelem;

	/** collision rate coefficient, [cm^3 s-1], upper to lower */
	float ColUL;

	/** optical depths - the escape probability is derived only from TauIn
	 * on first iteration.  on later iterations it is the average ofesc prob
	 * in inward (TauIn) and outward (TauTot - TauIn) directions */

	/** TauIn - the inward line optical depth [Napier], in the direction of the 
	 * continuum source.  
	 * This is measured from the illuminated face of 
	 * the cloud to the current position on the first iteration, and on second
	 * and later iterations in an open geometry or in an expanding closed geometry.
	 * For a static spherical geometry TauIn is the sum of the optical depth from the
	 * current position to the illuminated face plus the optical depth on the "other
	 * side".
	 * At the illuminated face in a static spherical geometry TauIn is TauTot / 2 */
	float TauIn;

	/** TauTot - total line optical depth [Napier] through the cloud. 
	 * TauTot is not used on the first iteration since it is not known.  
	 * On second and later iterations in an open or expanding closed geometry 
	 * this is the total optical depth through the computed structure.  
	 * For a static closed geometry this is twice the computed structure since the 
	 * geometry is assumed to be symmetric.
	 *
	 * when the double command is entered TauTot is set to twice the normal value
	 * to simulate the presence of material beyond the computed structure.
	 * */
	float TauTot;

	/** TauCon - line optical depth [Napier] to the continuum source from the
	 * illuminated face to the current position.
	 * For an open or expanding closed geometry TauCon is equal to TauIn.
	 * For a static closed geometry TauCon is optical depth from the illuminated face
	 * to the current depth */
	float TauCon;

	/** inward fraction [dimensionless] of total line emission*/
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float FracInwd;

	/** continuum pumping rate [s-1] from lower to upper level, A*occ num * g_up/g_lo,
	 * this is evaluated in RTMakeStat and RTMakeWind, which are called by HydroPEsc, RT_line_all */
	double pump;

	/** line intensity per unit time and vol [erg s-1 cm-3] */
	double xIntensity;

	/** number of photons emitted in the line per unit time and vol [photons s-1 cm-3] */
	double phots;

	/** gf value [dimensionless] */
	float gf;

	/** escape prob [dimensionless] */
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float Pesc;

	/** electron scattering escape prob [dimensionless] */
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float Pelec_esc;

	/** destruction probs [dimensionless] */
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float Pdest;

	/** damping constant is dampXvel divided by line width 
	 * units are velocity, since becomes dimensionless when div by line width in cm/2
	 * [cm s-1] */
	float dampXvel;

	/** [dimensionless] damping constant */
	float damp;

	/** cooling and heating due to collisional excitation [erg s-1 cm-3] */
	double cool , heat;

	/** [dimensionless] ratio of collisional to radiative excitation, C_lu / ( C_lu + pump )*/
	float ColOvTot;

	/** [dimensionless] collision strength or rates for transition */
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float cs;

	/** wavelentgh, usually in Angstroms, used for printout, can be any units */
	float WLAng;

	/** transition energy in degrees kelvin*/
	float EnergyK;

	/** transition energy in ergs */
	float EnergyErg;

	/** transition energy in wavenumbers */
	float EnergyWN;

	/** atomic constant part of line opacity per atom, 
	 * divide by line width in cm/s, DoppVel.doppler[nelem],
	 * to get line center opacity per atom, or absorption cross section, with units cm^2
	 * multiply by PopOpc to get PopOpc/dopper width, the true opacity (cm-1),
	 * then by length to get optical depth */
	float opacity;

	/** lower and upper statistical weights [dimensionless] */
	float gLo , gHi;

	/** populations of lower and upper levels [cm-3] */
	/** >>chng 01 nov 24, from float to double, need the precision for hyperfine structure lines */
	double PopLo , PopHi;

	/** population that enters net opacity after correction for stimulated emission [cm-3] */
	double PopOpc;

	/** transition prob, Einstein A upper to lower [s-1] */
	/** >>chng 06 mar 07, from double to float, save mem, no reason for double */
	float Aul;

	/** ots rate [cm-3 s-1] */
	double ots;

} EmLine;


/***************************************************************************
 * the following are prototypes for some routines that are part of the
 * debugging process - they come and go in any particular sub.  
 * it is not necessary to declare them when used since they are defined here
 **************************************************************************/

/**
 fudge enter fudge factors, or some arbitrary number, with fudge command
 \param ipnt
 */ 
double fudge(long int ipnt);

/**
 broken set flag saying that the code is broken 
 */ 
void broken(void);

/**fixit set flag saying that this code needs attention, but is not broken,
 * code is in service.cpp */
void fixit(void);

/**CodeReview - placed next to code that needs to be checked */ 
void CodeReview(void);

/**testcde set flag saying that test code is in place */
void TestCode(void);

/**MyMalloc wrapper for malloc().  Returns a good pointer or dies.
\param size use same type as library function malloc
\param file
\param line
*/
void *MyMalloc(size_t size, const char *file, int line);

/**MyCalloc wrapper for calloc().  Returns a good pointer or dies. 
\param num use same type as library function CALLOC
\param size
*/
void *MyCalloc(size_t num, size_t size);

/**MyRealloc wrapper for realloc().  Returns a good pointer or dies. 
\param num use same type as library function REALLOC
\param size
*/
void *MyRealloc(void *p, size_t size);

/** MyAssert a version of assert that fails gracefully
\param *file
\param line
*/ 
NORETURN void MyAssert(const char *file, int line);

/** the Cloudy exit handler */
NORETURN void cdEXIT(int);

/**path_not_set say that path has not been set */ 
void path_not_set( void );

/** print comment asking to show output to me */
void ShowMe(void);

/**TotalInsanity general error handler for something that cannot happen, exits */
NORETURN void TotalInsanity(void);

/**BadRead tried to read internal data and failed */
NORETURN void BadRead(void);

/**BadOpen tried to open data files and failed */
NORETURN void BadOpen(void);

/**
   NoNumb general error handler for no numbers on input line 
   \param *chCard
 */ 
NORETURN void NoNumb(char *chCard);

/** dbg_printf is a debug print routine that was provided by Peter Teuben,
 * as a component from his NEMO package.  It offers run-time specification
 * of the level of debugging */ 
void dbg_printf(int debug, const char *fmt, ...);


/***************************************************************************
 *
 * a series of Cloudy service routines, used throughout code,
 *
 **************************************************************************/

void cdDefines(void);

/**csphot returns photoionization cross section from opacity stage using std pointers 
\param inu INU is array index pointing to frequency where opacity is to be evaluated on f not c scale
\param ithr ITHR is pointer to threshold
\param iofset IOFSET is offset as defined in opac0
*/
double csphot(long int inu, long int ithr, long int iofset);

/** normal random variate generator 
\param xMean mean value
\param s standard deviation s
*/
double RandGauss(double xMean, double s );

/**A custom wrapper for RandGauss than truncates at two standard deviations. 
\param PctUncertainty
*/
double MyGaussRand( double PctUncertainty );

/**AnuUnit produce continuum energy in arbitrary units, ip is on C scale */
double AnuUnit(float energy);

/**cap4 convert first 4 char of input line chLab into chCAP all in caps, null termination 
\param chCAP output string, cap'd first 4 char of chLab,
\param chLab with null terminating input string ending with eol
*/ 
void cap4(char *chCAP , char *chLab);

/**caps convert input command line (through eol) to ALL CAPS 
\param chCard - line image as string of characters */
void caps(char *chCard );

/**e2 second exponential integral 
 \param t optical depth argument */
double e2(
	  double t );

/**ee1 first exponential integral
	\param x optical depth argument, returns e1(tau) */
double ee1(double x);

/**this one same as ee1, except is divided by a factor of exp(x), and is
 * only to be used for x>1.	
 \param x  optical depth argument, returns e1(tau) * exp(x) */
double ee1_safe(double x);

 /**FFmtRead - the free-format number reader
  \param *chCard string giving the line image
  \param *ipnt the index for the character in the string where we shall start
  \param last the number of characters in the string - do not search beyond it
  \param *lgEOL true if hit end of line with no number 
 */ 
double FFmtRead(const char *chCard, 
		long int *ipnt, 
		long int last, 
		bool *lgEOL);

 /**nMatch determine whether match to a keyword occurs on command line,
   return value is 0 if no match, and position of match within string if hit 
	  \param *chKey
	  \param *chCard
 */ 
long nMatch(const char *chKey, 
	    const char *chCard);

/**GetElem scans line image, finds element. returns atomic number j, 
   on C scale.  Returns , -1 if no hit.  chCARD_CAPS must have been
   converted to all CAPS by caps() to work 
\param chCARD_CAPS
*/ 
long int GetElem( char *chCARD_CAPS );

/** GetQuote get any name between double quotes off command line
 * return string as chLabel, is null terminated 
 * returns zero for success, 1 for did not find double quotes 
\param *chLabel we will generate a label and stuff it here
\param *chCard local cap'd line, we will blank this out
\param lgABORT if true then abort if no double quotes found, if false then
 return null string in this case
*/ 
int GetQuote(char *chLabel,	char *chCard, bool lgABORT );

/* want to define this only if no native os support exists */
#if !HAVE_POWI
/** powi raise x to integer power */
double powi( double , long int );
#endif

/** ipow
\return  m^n */
long int ipow( long, long );

/** print with 1p,e8.2 format onto stream FILE 
 * all are located in printe82.c */
void PrintE82( FILE*, double );

/** print with 1p,e8.1 format onto stream FILE */
void PrintE71( FILE*, double );

/** print with 1p,e9.3 format onto stream FILE */
void PrintE93( FILE*, double );

/** create string with val and format, to print with %s,
 * much faster than above, totally native on non-MS systems 
\param *fmt
\param val
*/
char *PrintEfmt(const char *fmt, double val );

/** this is -ln of smallest number sexp can handle */
const double SEXP_LIMIT = 84.;
/** sexp safe exponential function */
double sexp(double x);

/**dsexp safe exponential function for doubles 
\param x 
*/
 
double dsexp(double x);

/**plankf evaluate Planck function for any cell at current electron temperature 
\param ip
*/
 
double plankf(long int ip);

 /**			 
 32 point gaussian quadrature integration
 \param xl lower limit to integration
 \param xu - upper limit to integration
 \param (*fct) - pointer to routine to be integrated, arg is x val<BR>
 */ 
double qg32( double, double, double(*)(double) );
/* declar of optimize_func, the last arg, changed from double(*)() to above,
 * seemed to fix flags that were raised */


/**
   spsort netlib routine to sort array returning sorted indices
   \param x[] input array to be sorted	   
   \param  n number of values in x 	   
   \param  iperm[]  permutation output array
   \param  kflag flag saying what to do - 1 sorts into increasing order, not changing
   \param  kflag the original routine 
   \param  *ier error condition, should be 0
 */ 
void spsort( float x[], long int n, long int iperm[], int kflag, int *ier);


/**************************************************************************
 *
 * disable some bogus errors in the ms c compiler
 *
 **************************************************************************/

/* */
#ifdef _MSC_VER
	/* disable warning that conditional expression is constant, true or false in if */
#	pragma warning( disable : 4127 )
	/* disable strcat warning */
#	pragma warning( disable : 4996 )
	/* disable bogus underflow warning in MS VS*/
#	pragma warning( disable : 4056 )
	/* disable "inline function removed since not used", MS VS*/
#	pragma warning( disable : 4514 )
#endif
#ifdef __INTEL_COMPILER
#	pragma warning( disable : 1572 )/**< disable warning that floating-point comparisons are unreliable */
#endif
/* */

/*lint +e129 these resolve several issues pclint has with my system headers */
/*lint +e78 */
/*lint +e830 */
/*lint +e38 */
/*lint +e148 */
/*lint +e114 */
/*lint +e18 */
/*lint +e49 */
